# bash-yaml: functions to access yaml (config) files in bash
# Copyright (C)  2019 Heinz Deinhart
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# bash-yaml version 0.1.5
#  currently bash-yaml uses shyaml to parse files

#@@ Error Messages

#@ error-codes (bash-yaml)
# <50 .. reserved for base-base error-codes
#  51 .. illegal handle name
#  52 .. handle unknown
#  53 .. parent path not found
#  54 .. value name not found
#  9x .. errors local to each function

#@ byml_error_string <ret-vn> <error-code>
# returns string representation of error code
# errors: 2, 3
function byml_error_string() {
    (( $# == 2 )) || {
	bb_ierr "byml_error_string: called with $# instead of 2 args"
	return 2
    }

    local _ret_vn="$1"
    local _err_code="$2"

    bb_validate_vn bb_error_string "$_ret_vn" || return 3

    if (( "$_err_code" < 50 ))
    then
	bb_error_string "$_ret_vn" "$_err_code"
    else
	local _msg

	case "$_err_code" in
	    51) _msg="illegal handle name";;
	    52) _msg="handle unknown";;
	    53) _msg="parent path not found";;
	    54) _msg="value name not found";;
	    90|91|92|93|94|95|96|97|98|99) _msg="function local error";;
	    *) _msg="byml_error_string: unknown error code $_err_code";;
	esac
	printf -v "$_ret_vn" "%s" "$_msg"
    fi
}

#@@ Helper Functions

#@ byml_mapfun_escape_dots <dst-vn> <input-value>
# escapes all dots in input with backslashes
# errors: 2
function byml_mapfun_escape_dots() {
    (( $# == 2 )) || {
	bb_ierr "byml_mapfun_escape_dots: called with $# instead of 2 args"
	return 2
    }

    printf -v "$1" "%s" "${2//./\.}"
}

#@@ Handle handling

#@ byml_handle_is_open <handle>
# check if handle is open
# errors: 2, 4, 5
function byml_handle_is_open() {
    (( $# == 1 )) || {
	bb_ierr "byml_handle_is_open called with $# instead of 1 arg"
	return 2
    }

    local _yml_handle="$1"
    bb_validate_vn byml_handle_is_open "$_yml_handle" || return 51

    [[ -v "byml_handle_${_yml_handle}" ]] || return 52
}

#@ byml_handle_open <handle> <path>
# note: the current implementation does not read the file, it stores
#       only its name to be passed to shyaml later on
# errors: 2, 4, 8, 9
function byml_handle_open() {
    (( $# == 2 )) || {
	bb_ierr "byml_handle_open called with $# instead of 2 args"
	return 2
    }

    local _yml_handle="byml_handle_$1"
    local _inputfile="$2"

    bb_validate_vn byml_handle_open "$_yml_handle" || return 51

    [ -f "$_inputfile" ] || return 8
    [ -r "$_inputfile" ] || return 9

    declare -g "$_yml_handle"
    printf -v "$_yml_handle" "%s" "$_inputfile" 
}

#@ yml_handle_close <handle>
# errors: 2, 51, 52
function byml_handle_close() {
      (( $# == 1 )) || {
	bb_ierr "byml_handle_close called with $# instead of 1 arg"
	return 2
    }

    local _yml_handle="byml_handle_$1"

    bb_validate_vn byml_handle_open "$_yml_handle" || return 51

    [[ -v "${_yml_handle}" ]] || return 52

    unset "${_yml_handle}"
}

#@@ Getting Values

#@ byml_get_value <ret-vn> <handle> <parents-vn> <name>
# get a value from a well specified path
# errors: 2, 3, 52, 53, 54
function byml_get_value() {
    (( $# == 4 )) || {
	bb_ierr "byml_get_value: called with $# instead of 4 args"
	return 2
    }

    byml_handle_is_open "$2" || return 52

    local _ret_vn="$1"
    local _yml_handle="byml_handle_$2"
    local _parents_vn="$3"
    local _name="$4"

    bb_validate_vn byml_get_value "$_ret_vn" "$_parents_vn" || return 3

    # shellcheck disable=SC2034
    local byml_get_value__quoted_name
    byml_mapfun_escape_dots byml_get_value__quoted_name "$_name"

    local -a byml_get_value__quoted_parents
    local _byml_get_value__deli
    bb_array_map byml_get_value__quoted_parents "$_parents_vn" \
	byml_mapfun_escape_dots
    local byml_get_value__quoted_parent_string=
    if (( ${#byml_get_value__quoted_parents[@]} > 0 ))
    then
	bb_array_join_by byml_get_value__quoted_parent_string "." \
	    "${byml_get_value__quoted_parents[@]}"
	_byml_get_value__deli=.
    else
	# we are at top level: no parents, no delimiter
	_byml_get_value__deli=
    fi

    local _yml_file="${!_yml_handle}"

    # since currently handle is filename, check if file is readable
    [ -r "$_yml_file" ] || return 52

    local _byml_get_value__value

    if _byml_get_value__value=$(shyaml get-value \
	"$byml_get_value__quoted_parent_string$_byml_get_value__deli$_name" \
	< "$_yml_file" 2>/dev/null)
    then
	printf -v "${_ret_vn}" "%s" "$_byml_get_value__value"
    else
	if shyaml get-value "$byml_get_value__quoted_parent_string" \
	    < "$_yml_file" 2>/dev/null >/dev/null
	then
	    # no such value
	    return 54
	else
	    if (( ${#byml_get_value__quoted_parents[@]} > 0 ))
	    then
		# parent not found
		return 53
	    else
		# we are at top level
		return 54
	    fi
	fi
    fi
}

#@ byml_get_value_with_default <ret-vn> <yaml-handle> <parents-vn> <name> 
#                              <default-value>
# get a value from a well specified path
# takes default if value found but parent tree exists
# errors: 2, 3, 52, 53
function byml_get_value_with_default() {
    (( $# == 5 )) || {
	bb_ierr "byml_get_value_with_default: called with $# instead of 5 args"
	return 2
    }

    local _ret_vn="$1"
    local _yml_handle="$2"
    local _parents="$3"
    local _name="$4"
    local _defval="$5"

    bb_validate_vn byml_get_value_with_default "$_ret_vn" || return 3

    byml_get_value "$_ret_vn" "$_yml_handle" "$_parents" "$_name" || {
	local _i="$?"
	case "$_i" in
	    2|3|52|53) return "$_i";;
	    54) # parent path was found, return default
		printf -v "$_ret_vn" "%s" "$_defval";;
	    *) # that should not happen
		bb_msg warning "byml_get_value_with_default:" \
		    "internal error: byml_get_value returned: $_i"
		return 1;;
	esac
    }
}

#@ byml_get_value_with_parents <ret-vn> <yaml-handle> <parents-vn> <name>
# get a value from a path with fallbackup to parents
# where parent tree elements are seperated by '.'
# errors: 2, 3, 52, 53
function byml_get_value_with_parents() {
    (( $# == 4 )) || {
	bb_ierr "byml_get_value_with_parents: called with $# args instead of 4"
	return 2
    }

    local _ret_vn="$1"
    local _yml_handle="$2"
    local _parents_vn="$3"
    local _name="$4"

    bb_validate_vn byml_get_value_with_parents "$_ret_vn" "$_parents_vn" || \
	return 3

    # copy the array so we can modify it
    local -a byml_get_value_with_parents__parents
    bb_array_copy byml_get_value_with_parents__parents "$_parents_vn"

    local _retcode

    while true
    do
	if byml_get_value "$_ret_vn" "$_yml_handle" \
	    byml_get_value_with_parents__parents "$_name"
	then # we found a value
	    return 0
	else
	    # we did not found the value
	    _retcode="$?"
	fi

	case "$_retcode" in
	    2|3|52|53) return "$_retcode";;
	    54) # parent path was found
		(( ${#byml_get_value_with_parents__parents[@]} > 0 )) || \
		    return 54 # no parents left to check

		# try again at next parent level
		# shellcheck disable=SC2016
		unset 'byml_get_value_with_parents__parents[${#byml_get_value_with_parents__parents[@]}-1]'
		continue
		;;
	    *) # that should not happen
		bb_msg warning "byml_get_value_with_default:" \
		    "internal error: byml_get_value returned: $_retcode"
		return 1
		;;
	esac
	bb_ierr "byml_get_value_with_parents: reached unreachable code"
	return 1
    done
}

#@ byml_get_value_with_parents_and_default 
#   <ret-vn> <yaml-handle> <parents-vn> <name> <default-value>
# get a value from a well specified path, take default if not found
# errors: 1, 2, 6, 7, 8, 9
function byml_get_value_with_parents_and_default() {
    (( $# == 5 )) || {
	bb_ierr "byml_get_value_with_parents_and_default: called with $#" \
	    "instead of 5 args"
	return 2
    }

    local _ret_vn="$1"
    local _yml_handle="$2"
    local _parents="$3"
    local _name="$4"
    local _defval="$5"

    bb_validate_vn byml_get_value_with_parents_and_default "$_ret_vn" \
	"$_parents" || return 3

    byml_get_value_with_parents "$_ret_vn" "$_yml_handle" \
	"$_parents" "$_name" || {
	local _i="$?"
	case "$_i" in
	    2|3|52|53) return "$_i";;
	    54) # parent path was found, return default
		printf -v "$_ret_vn" "%s" "$_defval";;
	    *) # that should not happen
		bb_msg warning "byml_get_value_with_parents_and_default:" \
		    "internal error: byml_get_value_with_parents returned: $_i"
		return 1;;
	esac
    }
}

#@@ Lists and Dictionaries

# _byml_get_array <mode> <ret-vn> <yaml-handle> <value-name> <parents-vn>
# returns a list of keys of a struct as array
# errors: 2, 3, 52, 53, 91 (shyml parse error), 92 (illegal mode)
function _byml_get_array() {
    (( $# == 5 )) || {
	bb_ierr "byml_get_array: called with $# instead of 5 args"
	return 2
    }

    byml_handle_is_open "$3" || return 52

    local _mode_name="$1"
    local _ret_vn="$2"
    local _yml_handle="byml_handle_$3"
    local _value_name="$4"
    local _parents_vn="$5"

    local _mode
    case "$_mode_name" in
	dict_keys) _mode=keys;;
	dict_values) _mode=values;;
	list) _mode=get-values;;
	*)
	    bb_ierr "_byml_get_array: illegal mode: $_mode_name"
	    return 91
    esac

    bb_validate_vn byml_get_array "$_ret_vn" "$_parents_vn" || return 3

    # shellcheck disable=SC2034
    local byml_get_array__quoted_name
    byml_mapfun_escape_dots byml_get_array__quoted_name "$_value_name"

    local -a byml_get_array__quoted_parents 
    local byml_get_array__deli

    bb_array_map byml_get_array__quoted_parents "$_parents_vn" \
	byml_mapfun_escape_dots
    local byml_get_array__quoted_parent_string=
    if (( ${#byml_get_array__quoted_parents[@]} > 0 ))
    then
	bb_array_join_by byml_get_array__quoted_parent_string "." \
	    "${byml_get_array__quoted_parents[@]}"
	byml_get_array__deli=.
    else
	# we are at top level: no parents, no delimiter
	byml_get_array__deli=
    fi

    local _yml_file="${!_yml_handle}"

    # since currently handle is filename, check if file is readable
    [ -r "$_yml_file" ] || return 52

    local _shyaml_ref
    printf -v _shyaml_ref "%s%s%s" "$byml_get_array__quoted_parent_string" \
	"$byml_get_array__deli" "$_value_name"
    # do we work on whole file?

    local _shyaml_content

    # open async fd
    if [ -z "$_shyaml_ref" ]
    then
	# shellcheck disable=SC2093
	_shyaml_content="$(shyaml "$_mode" < "$_yml_file" 2>/dev/null)" || \
	    return 92 # should not happen at top level
    else
	local _shyaml_cmd
	printf -v "_shyaml_cmd" "%s%s%s" \
	    "$byml_get_array__quoted_parent_string" \
	    "$byml_get_array__deli" "$_value_name"
	_shyaml_content="$(shyaml "$_mode" "$_shyaml_cmd" < "$_yml_file" \
	    2>/dev/null)" || \
	    return 54 # no such value
    fi

    readarray -t "$_ret_vn" <<< "$_shyaml_content"
}

# _byml_get_array_with_parents <mode> <ret-vn> <yaml-handle> 
#                              <value-name> <parents-vn>
# returns all values of a list as bash array
# errors: 2, 3, 52, 53, 91 (shyml parse error)
function _byml_get_array_with_parents() {
    (( $# == 5 )) || {
 	bb_ierr "_byml_get_array_with_parents: called with $# instead of 5 args"
 	return 2
    }

    local _mode_name="$1"
    local _ret_vn="$2"
    local _yml_handle_name="$3"
    local _value_name="$4"
    local _parents_vn="$5"

    # copy the array so we can modify it
    local -a byml_get_array_with_parents__parents
    bb_array_copy byml_get_array_with_parents__parents "$_parents_vn"

    local _retcode

    while true
    do
	if _byml_get_array "$_mode_name" "$_ret_vn" "$_yml_handle_name" \
	    "$_value_name" byml_get_array_with_parents__parents
	then
	    return 0
	else
	    _retcode="$?"
	fi

	case "$_retcode" in
	    2|3|52|53|91) return "$_retcode";;
	    54) # parent path was found but value not
		(( ${#byml_get_array_with_parents__parents[@]} > 0 )) || \
		    return 54 # no parents left to check

		# try again at next parent level
		# shellcheck disable=SC2016
		unset 'byml_get_array_with_parents__parents[${#byml_get_array_with_parents__parents[@]}-1]'
		continue
		;;
	    *) # that should not happen
		bb_msg warning "_byml_get_array_with_parents:" \
		    "internal error: _byml_get_array failed with $_retcode"
		return 1
		;;
	esac
	bb_ierr "byml_get_array_with_parents: reached unreachable code"
	return 1
    done
}


#@ byml_get_dict_keys <ret-vn> <yaml-handle> <value-name> [<parents-vn>]
# returns all keys of a dictionary as bash array
# if value-name is "" then the whole yaml file is assumed to be a dict
# errors: 2, 3, 10, 12, 52, 53
function byml_get_dict_keys() {
    (( $# == 3 || $# == 4 )) || {
 	bb_ierr "byml_get_dict_keys: called with $# instead of 3 or 4 args"
 	return 2
    }

    # use empty parent list if non have been passed
    if (( $# == 4 ))
    then
	_byml_get_array dict_keys "$@"
    else
	# shellcheck disable=SC2034
	local -a byml_get_dict_keys__parents=()

	_byml_get_array dict_keys "$@" byml_get_dict_keys__parents
    fi

}

#@ byml_get_dict_keys_with_parents <ret-vn> <yaml-handle>
#                                  <value-name> <parents-vn>
# returns all keys of a dictionary as bash array
# search upwards through parents unitl value-name is found
# this will only work if the dict consists of scalar members only
# errors: 2, 3, 52, 53, 91 (shyml parse error)
function byml_get_dict_keys_with_parents() {
    (( $# == 4 )) || {
 	bb_ierr "byml_get_dict_keys_with_parents:" \
	    "called with $# instead of 3 args"
 	return 2
    }

    _byml_get_array_with_parents dict_keys "$@"
}

#@ byml_get_dict_values <ret-vn> <yaml-handle> <value-name> [<parents-vn>]
# returns all values of a dictionary as bash array
# this will only work if the dict consists of scalar members only
# if value-name is "" then the whole yaml file is assumed to be a dict
# errors: 2, 3, 10, 12, 52, 53
function byml_get_dict_values() {
    (( $# == 3 || $# == 4 )) || {
 	bb_ierr "byml_get_dict_values: called with $# instead of 3 or 4 args"
 	return 2
    }

    # use empty parent list if non have been passed
    if (( $# == 4 ))
    then
	_byml_get_array dict_values "$@"
    else
	# shellcheck disable=SC2034
	local -a byml_get_dict_values__parents=()

	_byml_get_array dict_values "$@" byml_get_dict_values__parents
    fi
}

#@ byml_get_dict_values_with_parents <ret-vn> <yaml-handle>
#                                    <value-name> <parents-vn>
# returns all values of a dictionary as bash array
# search upwards through parents unitl value-name is found
# this will only work if the dict consists of scalar members only
# errors: 2, 3, 52, 53, 91 (shyml parse error)
function byml_get_dict_values_with_parents() {
    (( $# == 4 )) || {
 	bb_ierr "byml_get_dict_values_with_parents:" \
	    "called with $# instead of 3 args"
 	return 2
    }

    _byml_get_array_with_parents dict_values "$@"
}

#@ byml_get_list <ret-vn> <yaml-handle> <list-name> [<parents-vn>]
# returns all values of a list as bash array
# this will only work if the array consists of scalar members only
# if value-name is "" then the whole yaml file is assumed to be an array
# errors: 2, 3, 10, 12, 52, 53
function byml_get_list() {
    (( $# == 3 || $# == 4 )) || {
 	bb_ierr "byml_get_list: called with $# instead of 3 or 4 args"
 	return 2
    }

    # use empty parent list if non have been passed
    if (( $# == 4 ))
    then
	_byml_get_array list "$@"
    else
	# shellcheck disable=SC2034
	local -a byml_get_list__parents=()

	_byml_get_array list "$@" byml_get_list__parents
    fi
}

#@ byml_get_list_with_parents <ret-vn> <yaml-handle> <value-name> <parents-vn>
# returns all values of a list as bash array
# search upwards through parents unitl value-name is found
# this will only work if the array consists of scalar members only
# errors: 2, 3, 52, 53, 91 (shyml parse error)
function byml_get_list_with_parents() {
    (( $# == 4 )) || {
 	bb_ierr "byml_get_list_with_parents: called with $# instead of 3 args"
 	return 2
    }

    _byml_get_array_with_parents list "$@"
}
